/*==========================================================================
**	
**	Filename	:	genTriangles.c
**	Date		:	12/Feb/2002
**	Author 		:	Rob Bateman
**	Notes		:	basic openGL program to generate some primitives
**	
**=========================================================================*/

/*==========================================================================
**	
**	Includes
**	
**=========================================================================*/

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>

#ifndef WIN32
#include <GL/glx.h>
#endif

/// the colour of the fog and the clear colour.
float  fog_colour[] = {0.6f,0.58f,0.79f,0.0f};
float  camera_pos[] = {0,0,0};

/// the density of the fog
float  fog_density  = 0.01f;

/// The fog mode, either GL_LINEAR, GL_EXP or GL_EXP2
GLenum fog_mode     = GL_EXP;

/// the far distance of the fog & camera
float far_distance  = 200.0f;

/// the near distance of the fog & camera
float near_distance = 0.05f;



#ifndef GL_EXT_fog_coord
#define GL_EXT_fog_coord 1

#define GL_FOG_COORDINATE_SOURCE_EXT      0x8450
#define GL_FOG_COORDINATE_EXT             0x8451
#define GL_FRAGMENT_DEPTH_EXT             0x8452
#define GL_CURRENT_FOG_COORDINATE_EXT     0x8453
#define GL_FOG_COORDINATE_ARRAY_TYPE_EXT  0x8454
#define GL_FOG_COORDINATE_ARRAY_STRIDE_EXT 0x8455
#define GL_FOG_COORDINATE_ARRAY_POINTER_EXT 0x8456
#define GL_FOG_COORDINATE_ARRAY_EXT       0x8457

typedef void (APIENTRY * PFNGLFOGCOORDFEXTPROC) (GLfloat coord);
typedef void (APIENTRY * PFNGLFOGCOORDFVEXTPROC) (const GLfloat *coord);
typedef void (APIENTRY * PFNGLFOGCOORDDEXTPROC) (GLdouble coord);
typedef void (APIENTRY * PFNGLFOGCOORDDVEXTPROC) (const GLdouble *coord);
typedef void (APIENTRY * PFNGLFOGCOORDPOINTEREXTPROC) (GLenum type, GLsizei stride, const GLvoid *pointer);
#endif

PFNGLFOGCOORDFEXTPROC glFogCoordfEXT = 0;
PFNGLFOGCOORDFVEXTPROC glFogCoordfvEXT = 0;
PFNGLFOGCOORDDEXTPROC glFogCoorddEXT = 0;
PFNGLFOGCOORDDVEXTPROC glFogCoorddvEXT = 0;
PFNGLFOGCOORDPOINTEREXTPROC glFogCoordPointerEXT = 0;



/*==========================================================================
**	
**	Global static variables
**	
**=========================================================================*/

static unsigned int g_dwDisplayListNum;
float zoom = 15.0f;
float rotx = 0;
float roty = 0.001f;
float tx = 0;
float ty = 0;
int lastx=0;
int lasty=0;
unsigned char Buttons[3] = {0};


/*==========================================================================
**	Func	:	generateCylinder
**	Params	:	iVsections	-	the number of sections up the cylinder
**				iHsections	-	the number of sections around the circumference
**				fHeight		-	the height of the cylinder
**				fRadius		-	the radius of the circular part	
**	Returns	:	a display list number
**	Notes	:	
**
**=========================================================================*/

void drawCylinder(  unsigned int iVsections,
					   		unsigned int iHsections,
					   		float        fHeight   ,
					   		float        fRadius,
							int x,
							int z)
{
	float fHalfHeight;
	float fHeightInc;
	float fAngle;
	unsigned int   i,j;
	float fog_c;
	float diff[3];
	diff[0] = camera_pos[0] - x;
	diff[2] = camera_pos[2] -z;
	diff[1] = camera_pos[1];
	fog_c = 1.3f*sqrt(diff[0]*diff[0]+diff[1]*diff[1]+diff[2]*diff[2]);
	
	/*
	**	make all params positive to prevent errors
	*/
	fHeight = fabs(fHeight);
	fRadius = fabs(fRadius);
	
	/*
	**	calculate some useful values
	*/
	fHalfHeight = 0.5f*fHeight;
	fHeightInc  = (float)fHeight/iHsections;

	/*
	**	enforce a sensible number of vertical sections
	*/
	if(iVsections < 5)
	{
		iVsections = 5;
	}
	

		/*
		**	Draw Top Circle as a triangle fan
		*/
		
		glBegin( GL_TRIANGLE_FAN );
			glFogCoordfEXT(fog_c);
			glNormal3f(0.0f,1.0f,0.0f);
			glVertex3f(0.0f,fHalfHeight,0.0f);
			for(i=0;i<=iVsections;i++)
			{
				fAngle = (float)i*(-6.2831853)/iVsections;
				glVertex3f(fRadius*cos(fAngle),fHalfHeight,fRadius*sin(fAngle));
			}
		
		glEnd();
		
		
		/*
		**	Draw Bottom Circle as a triangle fan
		*/
		
		glBegin( GL_TRIANGLE_FAN );
			glFogCoordfEXT(fog_c*2);
			glNormal3f(0.0f,-1.0f,0.0f);
			glVertex3f(0.0f,-fHalfHeight,0.0f);
			for(i=0;i<=iVsections;i++)
			{
				fAngle = (float)i*(6.2831853)/iVsections;
				glVertex3f(fRadius*cos(fAngle),-fHalfHeight,fRadius*sin(fAngle));
			}
		
		glEnd();
		
		
		
		/*
		**	Draw a number of sections that will make up the sides as quad strips
		*/

		for(i=0;i<iHsections;i++)
		{
			float fCurrMin = -fHalfHeight + fHeightInc*i,
				  fCurrMax = fCurrMin + fHeightInc;
			glBegin( GL_QUAD_STRIP );
				for(j=0;j<=iVsections;j++)
				{
					float ca,sa;
					fAngle = (float)j*(6.2831853)/iVsections;
					ca = fRadius*cos(fAngle);
					sa = fRadius*sin(fAngle);
					glNormal3f(ca,0.0f,sa);

					if(fCurrMin < -2)
						glFogCoordfEXT(fog_c*2);
					else
						glFogCoordfEXT(fog_c);
					glVertex3f(ca,fCurrMin,sa);
					if(fCurrMax < -2)
						glFogCoordfEXT(fog_c*2);
					else
						glFogCoordfEXT(fog_c);
					glVertex3f(ca,fCurrMax,sa);
				}			
			glEnd();
		}
	
	glEndList();
	
}



/*==========================================================================
**	Func	:	init
**	Params	:	
**	Returns	:	
**	Notes	:	Initialise Function to set default openGL parameters
**
**=========================================================================*/

void init( void )
{
	glClearColor(fog_colour[0],fog_colour[1],fog_colour[2],0.0f);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	/*
	**	Lighting setup
	*/
	{
		float fpAmbient[]  = {0.15f,0.15f,0.15f,0.0f};
		float fpDiffuse[]  = {0.8f,0.8f,0.8f,0.0f};
		float fpSpecular[] = {0.8f,0.8f,0.8f,0.0f};
		float fpPosition[] = {1.0f,16.0f,10.0f,1.0f};
		
		glLightfv(GL_LIGHT0,GL_AMBIENT,fpAmbient);
		glLightfv(GL_LIGHT0,GL_DIFFUSE,fpDiffuse);
		glLightfv(GL_LIGHT0,GL_SPECULAR,fpSpecular);
		glLightfv(GL_LIGHT0,GL_POSITION,fpPosition);
		
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0); 
	}
	
	/*
	**	Material Properties setup
	*/
	{
		float fpAmbient[]  = {0.3f,0.3f,0.3f,0.0f};
		float fpDiffuse[]  = {0.7f,0.7f,0.7f,0.0f};
		float fpSpecular[] = {1.0f,1.0f,1.0f,0.0f};
		
		glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,fpAmbient);
		glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,fpDiffuse);
		glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,fpSpecular);
		glMaterialf (GL_FRONT_AND_BACK,GL_SHININESS,35);
	}

	#ifdef WIN32
	glFogCoordfEXT       = (PFNGLFOGCOORDFEXTPROC)wglGetProcAddress("glFogCoordfEXT");
	glFogCoordfvEXT      = (PFNGLFOGCOORDFVEXTPROC)wglGetProcAddress("glFogCoordfvEXT");
	glFogCoorddEXT       = (PFNGLFOGCOORDDEXTPROC)wglGetProcAddress("glFogCoorddEXT");
	glFogCoorddvEXT      = (PFNGLFOGCOORDDVEXTPROC)wglGetProcAddress("glFogCoorddvEXT");
	glFogCoordPointerEXT = (PFNGLFOGCOORDPOINTEREXTPROC)wglGetProcAddress("glFogCoordPointerEXT");
	#else
	glFogCoordfEXT       = (PFNGLFOGCOORDFEXTPROC)glXGetProcAddressARB("glFogCoordfEXT");
	glFogCoordfvEXT      = (PFNGLFOGCOORDFVEXTPROC)glXGetProcAddressARB("glFogCoordfvEXT");
	glFogCoorddEXT       = (PFNGLFOGCOORDDEXTPROC)glXGetProcAddressARB("glFogCoorddEXT");
	glFogCoorddvEXT      = (PFNGLFOGCOORDDVEXTPROC)glXGetProcAddressARB("glFogCoorddvEXT");
	glFogCoordPointerEXT = (PFNGLFOGCOORDPOINTEREXTPROC)glXGetProcAddressARB("glFogCoordPointerEXT");
	#endif

	// set the fog attributes
	glFogf (GL_FOG_START,  near_distance);
	glFogf (GL_FOG_END,    far_distance);
	glFogfv(GL_FOG_COLOR,  fog_colour);      
	glFogi (GL_FOG_MODE,   fog_mode);
	glFogf (GL_FOG_DENSITY,fog_density);
	glFogi (GL_FOG_COORDINATE_SOURCE_EXT,GL_FOG_COORDINATE_EXT);

	// enable the fog
	glEnable(GL_FOG);

}



/*==========================================================================
**	Func	:	display
**	Params	:	
**	Returns	:
**	Notes	:	Display callback Function for screen re-draw
**
**=========================================================================*/

void display( void )
{
	int i,j;
	float mview_matrix[16];
	glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
	glLoadIdentity();

	glTranslatef(0,0,-zoom);
	glTranslatef(tx,ty,0);
	glRotatef(rotx,1,0,0);
	glRotatef(roty,0,1,0);

	glGetFloatv(GL_MODELVIEW_MATRIX,mview_matrix);

	camera_pos[0] = -mview_matrix[12];
	camera_pos[1] = -mview_matrix[13];
	camera_pos[2] = -mview_matrix[14];
	
	/* 
	**	draw geometry
	*/
	for(i=-20;i<=20;i+=4) {
		for(j=-20;j<=20;j+=4) {

			glPushMatrix();
				glTranslatef(i,3.3f,j);
				drawCylinder(12,20,6.6,0.5,i,j);
			glPopMatrix();
		}
	}

	glBegin( GL_LINES );
		glNormal3f(0,1,0);
		for(i=-20;i<=20;i+=2)
		{
			float fog_c;
			float diff[3];

			diff[0] = camera_pos[0] - i;
			diff[2] = camera_pos[2] + 20;
			diff[1] = camera_pos[1];
			fog_c = 1.3f*sqrt(diff[0]*diff[0]+diff[1]*diff[1]+diff[2]*diff[2]);
			glFogCoordfEXT(fog_c*2);
			glVertex3f(i,0,-20);

			diff[0] = camera_pos[0] - i;
			diff[2] = camera_pos[2] + 20;
			diff[1] = camera_pos[1];
			fog_c = 1.3f*sqrt(diff[0]*diff[0]+diff[1]*diff[1]+diff[2]*diff[2]);
			glFogCoordfEXT(fog_c*2);
			glVertex3f(i,0,20);

			diff[0] = camera_pos[0] - i;
			diff[2] = camera_pos[2] + 20;
			diff[1] = camera_pos[1];
			fog_c = 1.3f*sqrt(diff[0]*diff[0]+diff[1]*diff[1]+diff[2]*diff[2]);
			glFogCoordfEXT(fog_c*2);
			glVertex3f(-20,0,i);

			diff[0] = camera_pos[0] - i;
			diff[2] = camera_pos[2] + 20;
			diff[1] = camera_pos[1];
			fog_c = 1.3f*sqrt(diff[0]*diff[0]+diff[1]*diff[1]+diff[2]*diff[2]);
			glFogCoordfEXT(fog_c*2);
			glVertex3f(20,0,i);
		}

	glEnd();
	
	glutSwapBuffers();
}



/*==========================================================================
**	Func	:	reshape
**	Params	:	w	-	new screen width
**				h	-	new screen height
**	Returns	:	
**	Notes	:	Reshape callback function to resize gl display
**
**=========================================================================*/

void reshape( int w, int h )
{
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(52.0f,(h==0)?1.0f:(float)w/h,near_distance,far_distance);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}





//-------------------------------------------------------------------------------
//
void Motion(int x,int y)
{
	int diffx=x-lastx;
	int diffy=y-lasty;
	lastx=x;
	lasty=y;

	if( Buttons[0] && Buttons[1] )
	{
		zoom -= (float) 0.05f * diffx;
	}
	else
		if( Buttons[0] )
		{
			rotx += (float) 0.5f * diffy;
			roty += (float) 0.5f * diffx;		
		}
		else
			if( Buttons[1] )
			{
				tx += (float) 0.05f * diffx;
				ty -= (float) 0.05f * diffy;
			}
			glutPostRedisplay();
}

//-------------------------------------------------------------------------------
//
void Mouse(int b,int s,int x,int y)
{
	lastx=x;
	lasty=y;
	switch(b)
	{
	case GLUT_LEFT_BUTTON:
		Buttons[0] = ((GLUT_DOWN==s)?1:0);
		break;
	case GLUT_MIDDLE_BUTTON:
		Buttons[1] = ((GLUT_DOWN==s)?1:0);
		break;
	case GLUT_RIGHT_BUTTON:
		Buttons[2] = ((GLUT_DOWN==s)?1:0);
		break;
	default:
		break;		
	}
	glutPostRedisplay();
}



/*==========================================================================
**	Func	:	mouseClick
**	Params	:	key			-	id of button pressed
**				posX		-	x coord of screen where event happened
**				posY		-	y coord of screen where event happened
**	Returns	:	
**	Notes	:	mouse click callback function to register mouse presses
**
**=========================================================================*/

void keyboard(unsigned char key, int posX, int posY )
{
			
	glutPostRedisplay();
}

/*==========================================================================
**	Func	:	main
**	Params	:	argc	-	the number of command line arguments recieved
**				argv	-	an array of strings holding the arguments
**	Returns	:	
**	Notes	:	main function to set the ball rolling
**
**=========================================================================*/

int main( int argc, char** argv )
{
     glutInit(&argc, argv);
     glutInitDisplayMode   ( GLUT_DOUBLE |
                             GLUT_RGB    |
                             GLUT_DEPTH  );

     glutInitWindowSize    ( 800 , 600 );
     glutInitWindowPosition( 100 , 100 );

     glutCreateWindow( "Press 'c' to toggle object" );
     glutDisplayFunc(display);
     glutReshapeFunc(reshape);
	 glutMouseFunc(Mouse);
	 glutMotionFunc(Motion);
	 glutKeyboardFunc(keyboard);

     init();

	 glutMainLoop();

	 return 0;
}


	
	
